/* Copyright (C) 2002 Jean-Marc Valin 
   File: decode.c

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


   Portions Copyright 2002, John Edwards <john.edwards33@ntlworld.com>
 */

/* Takes a speex bitstream from input spx files and plays them through
   the soundcard using Windows Audio wave mapping, or decodes to file in
   8, 16, 24 or 32 bit integer, or 32 bit float format.  Multiple input 
   files are permitted, with wildcard use.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <windows.h>
#include <malloc.h>

#include "speex.h"
#include "speex_header.h"
#include "speex_stereo.h"
#include "speex_callbacks.h"
#include "ogg/ogg.h"

#include "audio.h"
#include "encode.h"
#include "misc.h"
#include "wave_out.h"           // included for WIN_AUDIO
#include "dither.h"             // included for dither functions

#define MAX_FRAME_SIZE 2000

#define ROUND64(x)   ( doubletmp = (x) + Dither.Add + (__int64)0x001FFFFD80000000L, *(__int64*)(&doubletmp) - (__int64)0x433FFFFD80000000L )

extern     int stop_decoding;

dither_t   Dither;
double     doubletmp;

/* Declare Functions */
static void usage(void);
double get_scale(oe_dec_opt *opt);

/* Dither output */
__int64 dither_output(short dithering, long i, double Sum, int k)
{
	short    shapingtype;
	double   Sum2;
	__int64  val;

	if(dithering){
		shapingtype = dithering - 1;
		if(!shapingtype){
			double  tmp = Random_Equi ( Dither.Dither );
			Sum2 = tmp - Dither.LastRandomNumber [k];
			Dither.LastRandomNumber [k] = tmp;
			Sum2 = Sum += Sum2;
			val = ROUND64 (Sum2)  &  Dither.Mask;
		}
		else{
			Sum2  = Random_Triangular ( Dither.Dither ) - scalar16 ( Dither.DitherHistory[k], Dither.FilterCoeff + i );
			Sum  += Dither.DitherHistory [k] [(-1-i)&15] = Sum2;
			Sum2  = Sum + scalar16 ( Dither.ErrorHistory [k], Dither.FilterCoeff + i );
			val = ROUND64 (Sum2)  &  Dither.Mask;
			Dither.ErrorHistory [k] [(-1-i)&15] = (float)(Sum - val);
		}
		return (val);
	}
	else
		return ((__int64)(ROUND64 (Sum)));
}

static void *process_header(ogg_packet *op, int enh_enabled, int *frame_size, int *rate, int *nframes, int forceMode,
                            int *channels, SpeexStereoState *stereo, int *extra_headers)
{
	void *st;
	SpeexMode *mode;
	SpeexHeader *header;
	int modeID;
	SpeexCallback callback;
      
	header = speex_packet_to_header((char*)op->packet, op->bytes);
	if (!header)
	{
		error_handler("Cannot read header\n");
		return NULL;
	}
	if (header->mode >= SPEEX_NB_MODES)
	{
		error_handler("Mode number %d does not (any longer) exist in this version\n", header->mode);
		return NULL;
	}
      
	modeID = header->mode;
	if (forceMode != -1)
		modeID = forceMode;
	mode = speex_mode_list[modeID];
   
	if (header->speex_version_id > 1)
	{
		error_handler("This file was encoded with Speex bit-stream version %d, which I don't know how to decode\n", header->speex_version_id);
		return NULL;
	}

	if (mode->bitstream_version < header->mode_bitstream_version)
	{
		error_handler("The file was encoded with a newer version of Speex. You need to upgrade in order to play it.\n");
		return NULL;
	}
	if (mode->bitstream_version > header->mode_bitstream_version) 
	{
		error_handler("The file was encoded with an older version of Speex. You would need to downgrade the version in order to play it.\n");
		return NULL;
	}
   
	st = speex_decoder_init(mode);
	if (!st) {
		error_handler("Decoder initialization failed.\n");
		return NULL;
   	}
	speex_decoder_ctl(st, SPEEX_SET_ENH, &enh_enabled);
	speex_decoder_ctl(st, SPEEX_GET_FRAME_SIZE, frame_size);

	if (!(*channels == 1)) {
		callback.callback_id = SPEEX_INBAND_STEREO;
		callback.func = speex_std_stereo_request_handler;
		callback.data = stereo;
		speex_decoder_ctl(st, SPEEX_SET_HANDLER, &callback);
	}
   
	if (!*rate)
		*rate = header->rate;
	/* Adjust rate if --force-* options are used */
	if (forceMode != -1)
	{
		if (header->mode < forceMode)
			*rate <<= (forceMode - header->mode);
		if (header->mode > forceMode)
			*rate >>= (header->mode - forceMode);
	}

	speex_decoder_ctl(st, SPEEX_SET_SAMPLING_RATE, rate);

	*nframes = header->frames_per_packet;

	if (*channels == -1)
		*channels = header->nb_channels;
   
/*	fprintf (stderr, "Decoding %d Hz audio using %s mode", *rate, mode->modeName);

	if (header->vbr)
		fprintf (stderr, " (VBR)\n");
	else
		fprintf(stderr, "\n");
*/
	*extra_headers = header->extra_headers;

	free(header);
	return st;
}

int oe_decode(oe_dec_opt *opt)
{
	ogg_sync_state    oy;
	ogg_page          og;
	ogg_packet        op;
	ogg_stream_state  os;

	FILE              *fin;
	audio_file        *fout;
	SpeexBits         bits;
	void              *sample_buffer,
	                  *st = NULL;
	double            bytesdone = 0,
	                  scale,
	                  dB;
	float             output[MAX_FRAME_SIZE];
	long              file_size;
	int               c,
	                  frame_size = 0,
	                  packet_count = 0,
	                  stream_init = 0,
	                  enh_enabled = 1,
	                  nframes = 2,
	                  print_bitrate = 0,
	                  close_in = 0,
	                  eos = 0,
	                  forceMode = -1,
	                  audio_size = 0,
	                  channels = -1,
	                  rate = 0,
	                  extra_headers;
	SpeexStereoState  stereo = SPEEX_STEREO_STATE_INIT;

	fin = fopen(opt->input_spx, "rb");
	if (!fin) {
		error_handler("Can't open input file: %s\n", opt->input_spx);
		return (0);
	}
	close_in = 1;

	fseek(fin, 0, SEEK_END);
	file_size = ftell(fin);
	rewind(fin);

	/*Init Ogg data struct*/
	ogg_sync_init(&oy);
   
	speex_bits_init(&bits);
	/*Main decoding loop*/
	while (1) {
		char *data;
		int i, j, k, nb_read;
		/*Get the ogg buffer for writing*/
		data = ogg_sync_buffer(&oy, 200);
		/*Read bitstream from input file*/
      		nb_read = fread(data, sizeof(char), 200, fin);      
		bytesdone += nb_read;

		opt->progress_update(file_size, bytesdone);
		ogg_sync_wrote(&oy, nb_read);

		/*Loop for all complete pages we got (most likely only one)*/
		while (ogg_sync_pageout(&oy, &og) == 1) {
			if (stream_init == 0) {
				ogg_stream_init(&os, ogg_page_serialno(&og));
				stream_init = 1;
			}
			/*Add page to the bitstream*/
			ogg_stream_pagein(&os, &og);
			/*Extract all available packets*/
			while (!eos && ogg_stream_packetout(&os, &op) == 1) {
				/*If first packet, process as Speex header*/
				if (packet_count == 0) {
					st = process_header(&op, enh_enabled, &frame_size, &rate, &nframes, forceMode, &channels, &stereo, &extra_headers);
					if (!nframes)
						nframes = 1;
					if (!st)
						return(0);
					if (!opt->decode_mode) {
						if (Set_WIN_Params (INVALID_FILEDESC, rate, SAMPLE_SIZE, channels) < 0) {
							error_handler("\nCan't access win32 play routines\n");
							return (0);
						}
					}
					else {
						fout = open_output_audio_file(opt->out_wave, rate, channels,
						              opt->format, OUTPUT_WAV);

						if (fout == NULL){
							error_handler("Not able to open output file: %s\n", opt->out_wave) ;
							return (0);
						}
					}
				}
				else if (packet_count == 1) {
//					print_comments((char*)op.packet, op.bytes);
//					error_handler ("File comments: ");
//					fwrite(op.packet, 1, op.bytes, stderr);
//					fprintf (stderr, "\n");
				}
				else if (packet_count <= 1+extra_headers) {
					/* Ignore extra headers */
				}
				else {
					/*End of stream condition*/
					if (op.e_o_s)
						eos = 1;

					/*Copy Ogg packet to Speex bitstream*/
					speex_bits_read_from(&bits, (char*)op.packet, op.bytes);
					for (j = 0; j < nframes; j++) {
						sample_buffer = malloc(sizeof(float)*channels*frame_size);
						/*Decode frame*/
						speex_decode(st, &bits, output);

						if (speex_bits_remaining(&bits) < 0) {
							error_handler("\nDecoding overflow: corrupted stream?\n");
							break;
						}
						if(channels == 2)
							speex_decode_stereo(output, frame_size, &stereo);

						/*PCM saturation (just in case)*/
						for (i = 0, k = 0; i < frame_size*channels; i++, k++) {
							__int64 val;
							double Sum;
							output[i] /= 32767;
							if(opt->man_gain)
								output[i] *= pow(10., (double)opt->man_gain/20);
							if(opt->limiter){	/* hard 6dB limiting */
								if (output[i] < -0.5)
									output[i] = tanh((output[i] + 0.5) / (1-0.5)) * (1-0.5) - 0.5;
								else if (output[i] > 0.5)
									output[i] = tanh((output[i] - 0.5) / (1-0.5)) * (1-0.5) + 0.5;
							}
							if(opt->format != WAV_FMT_FLOAT){
								Sum = output[i] * 2147483647.f;
								if(k > 31)
									k = 0;
								val = dither_output(opt->dither, k, Sum, 1) / opt->conv_factor;
								if (val > (__int64)opt->wrap_prev_pos)
									val = (__int64)opt->wrap_prev_pos;
								else if (val < (__int64)opt->wrap_prev_neg)
									val = (__int64)opt->wrap_prev_neg;
								output[i] = (float)val;
							}
							else{
								if (output[i] > opt->wrap_prev_pos)
									output[i] = opt->wrap_prev_pos;
								else if (output[i] < opt->wrap_prev_neg)
									output[i] = opt->wrap_prev_neg;
							}
						}
						sample_buffer = output_to_PCM(output, sample_buffer, channels,
						                frame_size, opt->format);
						if(!opt->decode_mode)         // playback
							WIN_Play_Samples (sample_buffer, sizeof(short) * channels * frame_size);
						else
							write_audio_file(fout, sample_buffer, frame_size * channels);
						free(sample_buffer);
					}
				}
				packet_count++;
			}
		}
		if(stop_decoding)
			break;
		if (feof(fin))
			break;
	}

	if (st)
		speex_decoder_destroy(st);
	speex_bits_destroy(&bits);
	ogg_sync_clear(&oy);
	ogg_stream_clear(&os);

	/* Finished with this file, so cleanup. */
	if(opt->decode_mode == 0)
		WIN_Audio_close();
	else
		close_audio_file(fout);

	if (opt->del_input)
		if (remove(opt->input_spx) != 0)
			error_handler("Error deleting input file\n");

	return (1);
}

/******************************** end of decode.c ********************************/

